package springbatchdemo;

import org.springframework.batch.core.*;
import org.springframework.batch.core.configuration.annotation.*;
import org.springframework.batch.core.launch.support.RunIdIncrementer;
import org.springframework.batch.core.scope.context.*;
import org.springframework.batch.core.step.tasklet.Tasklet;
import org.springframework.batch.repeat.RepeatStatus;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.*;

@Configuration
@EnableBatchProcessing
public class ConditionalFlowJobConfiguration
{
   public static final String OK_ODER_FEHLER = "OK_ODER_FEHLER";
   @Autowired private JobBuilderFactory  jobBuilderFactory;
   @Autowired private StepBuilderFactory stepBuilderFactory;

   /** Tasklet: Zeigt Text im Kommandozeilenfenster */
   public class PrintTextTasklet implements Tasklet
   {
      final String text;

      public PrintTextTasklet( String text ) { this.text = text; }

      @Override
      public RepeatStatus execute( StepContribution sc, ChunkContext cc ) throws Exception {
         System.out.println( text );
         return RepeatStatus.FINISHED;
      }
   }

   /** Tasklet: "Geschaeftsprozess", entweder erfolgreich oder mit Fehler */
   public class ArbeitsTasklet extends PrintTextTasklet
   {
      public ArbeitsTasklet( String text ) { super( text ); }

      @Override
      public RepeatStatus execute( StepContribution sc, ChunkContext cc ) throws Exception {
         StepContext stpCtx = cc.getStepContext();
         System.out.println( "\n---- Job: " + stpCtx.getJobName() + ", mit JobParametern: " + stpCtx.getJobParameters() );
         String okOderFehler = stpCtx.getStepExecution().getJobParameters().getString( OK_ODER_FEHLER );
         if( ( okOderFehler != null ) ? !okOderFehler.equalsIgnoreCase( "ok" ) : (Math.random() < 0.5) ) {
            System.out.println( super.text + ": mit Fehler" );
            throw new Exception( "-- Dieser Fehler ist korrekt! --" );
         }
         System.out.println( super.text + ": ok" );
         return RepeatStatus.FINISHED;
      }
   }

   @Bean
   public Step arbeitsStep()
   {
      return stepBuilderFactory.get( "arbeitsStep" ).tasklet( new ArbeitsTasklet( "ArbeitsStep" ) ).build();
   }

   @Bean
   public Step fehlerbehandlungsStep()
   {
      return stepBuilderFactory.get( "fehlerbehandlungsStep" ).tasklet( new PrintTextTasklet( "FehlerbehandlungsStep" ) ).build();
   }

   @Bean
   public Step okStep()
   {
      return stepBuilderFactory.get( "okStep" ).tasklet( new PrintTextTasklet( "OkStep" ) ).build();
   }

   @Bean
   public Step abschliessenderStep()
   {
      return stepBuilderFactory.get( "abschliessenderStep" ).tasklet( new PrintTextTasklet( "AbschliessenderStep" ) ).build();
   }

   /** Job Abc: fuehrt Steps im "Conditional Flow" aus */
   @Bean
   public Job meinConditionalFlowJobAbc()
   {
      return jobBuilderFactory.get( "jobAbc" ).incrementer( new RunIdIncrementer() )
              .flow( arbeitsStep() ).on( ExitStatus.FAILED.getExitCode() ).to( fehlerbehandlungsStep() ).next( abschliessenderStep() )
              .from( arbeitsStep() ).on( "*" ).to( okStep() ).next( abschliessenderStep() )
              .end().build();
   }
}
