package itemchunkdemo;

import javax.sql.DataSource;
import org.springframework.batch.core.*;
import org.springframework.batch.core.configuration.annotation.*;
import org.springframework.batch.core.launch.support.RunIdIncrementer;
import org.springframework.batch.item.database.BeanPropertyItemSqlParameterSourceProvider;
import org.springframework.batch.item.database.JdbcBatchItemWriter;
import org.springframework.batch.item.file.FlatFileItemReader;
import org.springframework.batch.item.file.mapping.*;
import org.springframework.batch.item.file.transform.DelimitedLineTokenizer;
import org.springframework.beans.factory.annotation.*;
import org.springframework.context.annotation.*;
import org.springframework.core.io.*;
import org.springframework.jdbc.core.JdbcTemplate;

/** Definition des Batch-Prozesses */
@Configuration
@EnableBatchProcessing
public class ItemChunkJobConfiguration
{
   @Autowired public JobBuilderFactory  jobBuilderFactory;
   @Autowired public StepBuilderFactory stepBuilderFactory;
   @Autowired public DataSource dataSource;

   /** ItemReader: liest Input-Daten, im Beispiel aus CSV-Datei */
   @Bean
   @StepScope
   public FlatFileItemReader<Meinedaten> meinItemReader( @Value("#{jobParameters['InputDateiPfad']}") String inputDateiPfad )
   {
      FlatFileItemReader<Meinedaten> reader = new FlatFileItemReader<Meinedaten>();
      reader.setResource( ( inputDateiPfad != null )
                          ? new FileSystemResource( inputDateiPfad )
                          : new ClassPathResource( "BeispielDaten1.csv" ) );
      reader.setLineMapper( new DefaultLineMapper<Meinedaten>() {{
         setLineTokenizer( new DelimitedLineTokenizer() {{
            setNames( new String[] { "Vorname", "Nachname" } );
         }} );
         setFieldSetMapper( new BeanWrapperFieldSetMapper<Meinedaten>() {{
            setTargetType( Meinedaten.class );
         }} );
      }} );
      return reader;
   }

   /** ItemProcessor: fuehrt Geschaeftslogik aus */
   @Bean
   public MeinedatenItemProcessor meinItemProcessor()
   {
      return new MeinedatenItemProcessor();
   }

   /** ItemWriter: schreibt Output-Daten, im Beispiel in Datenbank */
   @Bean
   public JdbcBatchItemWriter<Meinedaten> meinItemWriter()
   {
      JdbcBatchItemWriter<Meinedaten> writer = new JdbcBatchItemWriter<Meinedaten>();
      writer.setItemSqlParameterSourceProvider( new BeanPropertyItemSqlParameterSourceProvider<Meinedaten>() );
      writer.setSql( "Insert into Meinedaten ( Vorname, Nachname, Buchstabenanzahl ) Values ( :vorname, :nachname, :buchstabenanzahl )" );
      writer.setDataSource( dataSource );
      return writer;
   }

   /** Step: bearbeitet Chunks von Items, und ruft hierzu ItemReader, ItemProcessor und ItemWriter auf */
   @Bean
   public Step meinStep()
   {
      return stepBuilderFactory.get( "meinStep" )
              .<Meinedaten,Meinedaten> chunk( 10 )
              .reader(    meinItemReader( null ) )
              .processor( meinItemProcessor() )
              .writer(    meinItemWriter() )
              .build();
   }

   /** Job: fuehrt Steps aus */
   @Bean
   public Job meinItemChunkJob()
   {
      return jobBuilderFactory.get( "meinItemChunkJob" )
              .incrementer( new RunIdIncrementer() )
              .listener( meinListener() )
              .flow(     meinStep() )
              .end()
              .build();
   }

   /** Completion-Notification-Listener: ueberprueft das Ergebnis */
   @Bean
   public JobExecutionListener meinListener()
   {
      return new JobCompletionNotificationListener( new JdbcTemplate( dataSource ) );
   }
}
