package springbatchdemo;

import static org.hamcrest.Matchers.is;
import static org.junit.Assert.*;
import java.util.Map;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.batch.core.*;
import org.springframework.batch.core.explore.JobExplorer;
import org.springframework.batch.core.launch.JobOperator;
import org.springframework.batch.item.ExecutionContext;
import org.springframework.beans.factory.annotation.*;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

@RunWith( SpringJUnit4ClassRunner.class )
@ContextConfiguration( classes = MeineApplication.class )
public class StoppableDelayedRetryJobTest
{
   @Autowired private JobExplorer jobExplorer;
   @Autowired @Qualifier( "asyncJobOperator" ) private JobOperator asyncJobOperator;

   @Test
   public void testWiederholungMitWartezeit() throws JobExecutionException
   {
      // OK-Fall:
      //  1.: AvoidDuplicateRun,
      //  2.: ErstesTasklet: ok,
      //  3.: ZweitesTasklet: ok,
      //  4.: OkStep,
      //  5.: AbschliessenderStep
      Long executionId1 = asyncJobOperator.start( StoppableDelayedRetryJobConfiguration.JOB_NAME, "OK_ODER_FEHLER=ok" );
      JobExecution je1 = warteBisStatus( executionId1, BatchStatus.COMPLETED, 10 );
      printStepExecutions( je1 );
      assertThat( "Job-Name falsch", je1.getJobInstance().getJobName(), is( StoppableDelayedRetryJobConfiguration.JOB_NAME ) );
      assertThat( "Job-Step-Anzahl falsch",         je1.getStepExecutions().size(), is( 5 ) );
      assertThat( "FailureExceptions",              je1.getAllFailureExceptions().size(), is( 0 ) );
      assertThat( "Job ist nicht COMPLETED",        je1.getStatus(), is( BatchStatus.COMPLETED ) );
      assertThat( "ExitStatus ist nicht COMPLETED", je1.getExitStatus().getExitCode(), is( ExitStatus.COMPLETED.getExitCode() ) );
      assertThat( je1.getExitStatus().getExitDescription(), is( StoppableDelayedRetryJobConfiguration.ERGEBNIS_OK ) );
      assertThat( "" + je1.getExecutionContext(), is( "{Msg=ErstesTasklet, ZweitesTasklet, OkStep, AbschliessenderStep}" ) );

      // Fehler-Fall, nach 4 erfolglosen Versuchen (= 9 Steps) wird mit Fehler abgebrochen:
      //  1.: AvoidDuplicateRun,
      //  2.: ErstesTasklet: mit Fehler
      //  3.: WarteTasklet: Wiederholung nach Wartezeit . . . . . . . . . .
      //  4.: ErstesTasklet: mit Fehler
      //  5.: WarteTasklet: Wiederholung nach Wartezeit . . . . . . . . . .
      //  6.: ErstesTasklet: mit Fehler
      //  7.: WarteTasklet: Wiederholung nach Wartezeit . . . . . . . . . .
      //  8.: ErstesTasklet: mit Fehler
      //  9.: WarteTasklet: keine Wiederholung
      // 10.: FehlerbehandlungsStep
      // 11.: AbschliessenderStep
      Long executionId2 = asyncJobOperator.start( StoppableDelayedRetryJobConfiguration.JOB_NAME,
                                                  "OK_ODER_FEHLER=Fehler,ANZ_WIEDERHOL=7" );
      JobExecution je2 = warteBisStatus( executionId2, BatchStatus.COMPLETED, 10 );
      printStepExecutions( je2 );
      assertThat( "Job-Name falsch", je2.getJobInstance().getJobName(), is( StoppableDelayedRetryJobConfiguration.JOB_NAME ) );
      assertThat( "Job-Step-Anzahl falsch",         je2.getStepExecutions().size(), is( 11 ) );
      assertThat( "FailureExceptions",              je2.getAllFailureExceptions().size(), is( 0 ) );
      assertThat( "Job ist nicht COMPLETED",        je2.getStatus(), is( BatchStatus.COMPLETED ) );
      assertThat( "ExitStatus ist nicht COMPLETED", je2.getExitStatus().getExitCode(), is( ExitStatus.COMPLETED.getExitCode() ) );
      assertThat( je2.getExitStatus().getExitDescription(), is( StoppableDelayedRetryJobConfiguration.ERGEBNIS_FEHLER ) );

      // 2 Versuche (5 Steps) mit Fehler, dann erfolgreich weiter mit OK:
      //  1.: AvoidDuplicateRun,
      //  2.: ErstesTasklet: mit Fehler
      //  3.: WarteTasklet: Wiederholung nach Wartezeit . . . . . . . . . .
      //  4.: ErstesTasklet: mit Fehler
      //  5.: WarteTasklet: Wiederholung nach Wartezeit . . . . . . . . . .
      //  6.: ErstesTasklet: auf ok gesetzt, weil anzahlStepExecutions(=6) > anzStepFehlOk(=5)
      //  7.: ZweitesTasklet
      //  8.: OkStep
      //  9.: AbschliessenderStep
      Long executionId3 = asyncJobOperator.start( StoppableDelayedRetryJobConfiguration.JOB_NAME,
                                                  "OK_ODER_FEHLER=Fehler,ANZ_WIEDERHOL=7,ANZ_STEP_FEHL_OK=5" );
      JobExecution je3 = warteBisStatus( executionId3, BatchStatus.COMPLETED, 10 );
      printStepExecutions( je3 );
      assertThat( "Job-Name falsch", je3.getJobInstance().getJobName(), is( StoppableDelayedRetryJobConfiguration.JOB_NAME ) );
      assertThat( "Job-Step-Anzahl falsch",         je3.getStepExecutions().size(), is( 9 ) );
      assertThat( "FailureExceptions",              je3.getAllFailureExceptions().size(), is( 0 ) );
      assertThat( "Job ist nicht COMPLETED",        je3.getStatus(), is( BatchStatus.COMPLETED ) );
      assertThat( "ExitStatus ist nicht COMPLETED", je3.getExitStatus().getExitCode(), is( ExitStatus.COMPLETED.getExitCode() ) );
      assertThat( je3.getExitStatus().getExitDescription(), is( StoppableDelayedRetryJobConfiguration.ERGEBNIS_OK ) );
   }

   // Stoppable-Test: Stopp-Signal nach 0,5 Sekunden, waehrend der Wartezeit im WarteTasklet:
   // -->
   //  1.: AvoidDuplicateRun,
   //  2.: ErstesTasklet: mit Fehler
   //  3.: WarteTasklet: Wiederholung nach Wartezeit . . . . .  !!! Stopp-Signal durch StoppableTasklet !!!
   // -->
   // StepExecution 0: CommitCount = 1, Status = COMPLETED, ExitStatus = COMPLETED, StepName = avoidDuplicateRun
   // StepExecution 1: CommitCount = 1, Status = COMPLETED, ExitStatus = FAILED,    StepName = ersterStep
   // StepExecution 2: CommitCount = 1, Status = STOPPED,   ExitStatus = STOPPED,   StepName = warteStep, ... Stopp-Signal ...
   @Test
   public void testStoppable() throws JobExecutionException
   {
      final long wartezeitBisAbbruch = 500;

      System.out.println( "\n---- Vorhandene Jobs: " + asyncJobOperator.getJobNames() );

      Long executionId = asyncJobOperator.start( StoppableDelayedRetryJobConfiguration.JOB_NAME, "Test=Stop,OK_ODER_FEHLER=Fehler" );

      try { Thread.sleep( wartezeitBisAbbruch ); } catch( InterruptedException e ) {}

      JobExecution je = jobExplorer.getJobExecution( executionId );
      assertThat( je.getStatus(), is( BatchStatus.STARTED ) );

      asyncJobOperator.stop( executionId.longValue() );

      je = warteBisStatus( executionId, BatchStatus.STOPPED, 2 );
      printStepExecutions( je );
      assertThat( "Job-Name falsch", je.getJobInstance().getJobName(), is( StoppableDelayedRetryJobConfiguration.JOB_NAME ) );
      assertThat( "Job-Step-Anzahl falsch",       je.getStepExecutions().size(), is( 3 ) );
      assertThat( "FailureExceptions",            je.getAllFailureExceptions().size(), is( 0 ) );
      assertThat( "Job ist nicht STOPPED",        je.getStatus(), is( BatchStatus.STOPPED ) );
      assertThat( "ExitStatus ist nicht STOPPED", je.getExitStatus().getExitCode(), is( ExitStatus.STOPPED.getExitCode() ) );
      assertThat( je.getExitStatus().getExitDescription(),
                  is( "org.springframework.batch.core.JobInterruptedException; " +
                      StoppableDelayedRetryJobConfiguration.STOPP_SIGNAL_DURCH_STOPPABLETASKLET ) );
      assertThat( "" + je.getExecutionContext(), is( "{Msg=ErstesTasklet, WarteTasklet}" ) );
      long dauerJob = je.getEndTime().getTime() - je.getStartTime().getTime();
      System.out.println( "Jobdauer: " + dauerJob + " ms" );
      assertTrue( dauerJob < wartezeitBisAbbruch * 15 / 10 );
   }

   // Doppelter Job-Start:
   // -->
   //  !!! Abbruch weil Job bereits laeuft !!!
   // StepExecution: Status = COMPLETED, ExitStatus = FAILED, StepName = avoidDuplicateRun, ExitDescription = Abbruch weil Job bereits laeuft
   // JobExecution:  Status = FAILED,    ExitStatus = FAILED, ExitDescription = Abbruch weil Job bereits laeuft
   @Test
   public void testDuplicateRun() throws JobExecutionException
   {
      Long executionId1 = asyncJobOperator.start( StoppableDelayedRetryJobConfiguration.JOB_NAME,
                                                  "Test=Dupl1,OK_ODER_FEHLER=Fehler,ANZ_STEP_FEHL_OK=4" );

      try { Thread.sleep( 500 ); } catch( InterruptedException e ) {}

      Long executionId2 = asyncJobOperator.start( StoppableDelayedRetryJobConfiguration.JOB_NAME,
                                                  "Test=Dupl2,OK_ODER_FEHLER=Fehler,ANZ_STEP_FEHL_OK=4" );

      JobExecution je1 = warteBisStatus( executionId1, BatchStatus.COMPLETED, 10 );
      JobExecution je2 = warteBisStatus( executionId2, BatchStatus.FAILED,    10 );

      System.out.println( "\n1. Job:" );
      printStepExecutions( je1 );
      assertThat( "Job-Step-Anzahl falsch",         je1.getStepExecutions().size(), is( 9 ) );
      assertThat( "FailureExceptions",              je1.getAllFailureExceptions().size(), is( 0 ) );
      assertThat( "Job ist nicht COMPLETED",        je1.getStatus(), is( BatchStatus.COMPLETED ) );
      assertThat( "ExitStatus ist nicht COMPLETED", je1.getExitStatus().getExitCode(), is( ExitStatus.COMPLETED.getExitCode() ) );
      assertThat( je1.getExitStatus().getExitDescription(), is( StoppableDelayedRetryJobConfiguration.ERGEBNIS_OK ) );

      System.out.println( "\n2. Job:" );
      printStepExecutions( je2 );
      assertThat( "Job-Step-Anzahl falsch",         je2.getStepExecutions().size(), is( 1 ) );
      assertThat( "FailureExceptions",              je2.getAllFailureExceptions().size(), is( 0 ) );
      assertThat( "Job ist nicht FAILED",           je2.getStatus(), is( BatchStatus.FAILED ) );
      assertThat( "ExitStatus ist nicht FAILED",    je2.getExitStatus().getExitCode(), is( ExitStatus.FAILED.getExitCode() ) );
      assertThat( je2.getExitStatus().getExitDescription(), is( StoppableDelayedRetryJobConfiguration.ABBRUCH_WEIL_JOB_BEREITS_LAEUFT ) );
      assertThat( "" + je2.getExecutionContext(), is( "{}" ) );
   }

   private JobExecution warteBisStatus( Long executionId, BatchStatus stat, int maxWartezeitSek )
   {
      for( int i = 0; i < 5 * maxWartezeitSek; i++ ) {
         JobExecution je = jobExplorer.getJobExecution( executionId );
         if( stat == null || je == null || stat.equals( je.getStatus() ) ) {
            return je;
         }
         try { Thread.sleep( 200 ); } catch( InterruptedException e ) {}
      }
      return jobExplorer.getJobExecution( executionId );
   }

   private static void printStepExecutions( JobExecution je )
   {
      for( StepExecution se : je.getStepExecutions() ) {
         String s = se.getExitStatus().getExitDescription();
         if( s != null && s.length() > 0 ) { s = ", ExitDescription = " + s; }
         System.out.println( "StepExecution " + se.getId() + ": CommitCount = " + se.getCommitCount() +
               ", Status = " + se.getStatus() + ", ExitStatus = " + se.getExitStatus().getExitCode() +
               ", StepName = " + se.getStepName() + s );
      }
      String s = je.getExitStatus().getExitDescription();
      if( s != null && s.length() > 0 ) { s = ", ExitDescription = " + s; }
      System.out.println( "JobExecution  " + je.getId() + ": JobId = " + je.getJobId() +
            ",       Status = " + je.getStatus() + ", ExitStatus = " + je.getExitStatus().getExitCode() + s );
      ExecutionContext ec = je.getExecutionContext();
      if( ec != null && !ec.isEmpty() ) {
         for( Map.Entry<String, Object> e : ec.entrySet() ) {
            System.out.println( "Job-ExecutionContext: " + e.getKey() + ": " + e.getValue() );
         }
      }
   }
}
